import exceptions
import time
from selenium.webdriver.common.keys import Keys
from selenium import webdriver
import unittest


class TestHarkDesigner(unittest.TestCase):
    """ A basic class for unittesting HARK Designer.
    Inherit this class to make your test.

    """
    def configure(self):
        """ initializes browser and url.
        In default,
            browser = "Chrome" and
            url="http://localhost:3000/?selenium=selenium".
        browser can be Chrome or Firefox or Safari """
        self.browser = "Chrome"
        self.url = "http://localhost:3000/?selenium=selenium"
        self.delim = "_delim_"
        self.timeout = 3

    def setUp(self):
        """ open driver and configure. This method is called automatically."""
        self.configure()
        if self.browser == "Chrome":
            self.driver = webdriver.Chrome()
        elif self.browser == "Firefox":
            self.driver = webdriver.Firefox()
        elif self.browser == "Safari":
            self.driver = webdriver.Safari()
        else:
            raise exceptions.ValueError(
                "browser must be Chrome/Firefox/Safari")

        self.driver.implicitly_wait(self.timeout)
        self.driver.get(self.url)
        # self.driver.maximize_window()
        time.sleep(2)
        self.zoom(3)

    def tearDown(self):
        """Finalization of the test. This is automatically run by unittest"""
        if self.driver:
            self.driver.close()
            self.driver.quit()

    def connect(self, frId, frTerm, toId, toTerm):
        """ connect the link
        frId, frTerm  --> toId, toTerm """
        frSelector = "//div[@id='%s%s%s']/div" % (frId, self.delim, frTerm)
        toSelector = "//div[@id='%s%s%s']/div" % (toId, self.delim, toTerm)

        # test node
        self.assertNodeExists(frId)
        self.assertNodeExists(toId)

        # testParameter
        self.assertTermExists(frId, frTerm)
        self.assertTermExists(toId, toTerm)

        # get element objects
        frNodeElem = self.driver.find_element_by_id(frId)
        frElem = self.driver.find_element_by_xpath(frSelector)
        toElem = self.driver.find_element_by_xpath(toSelector)

        # check for the connection direction
        frSide = frElem.find_element_by_xpath("../..").get_attribute("class")
        if "interminals" in frSide:
            offset = -2
        elif "outterminals" in frSide:
            offset = 2
        else:
            raise exceptions.ValueError(
                "connection failed. the terminal has wrong class: " + frSide)

        # connect
        action = webdriver.ActionChains(self.driver)
        action.click_and_hold(frElem)
        for i in range(5):
            action.move_by_offset(offset, 0)
        action.move_to_element(toElem)
        action.release(toElem)
        action.perform()
        time.sleep(1.0)

    def addNode(self, name):
        """Add a node."""
        elem = self.driver.find_element_by_css_selector(
            "#nodelistSearch input")
        elem.send_keys(name)
        nodes = self.driver.execute_script(
            'return _.map($("#flatNodelist li:visible"), '
            'function(n){return $(n).attr("key");})')
        self.assertIn(name, nodes,
                      msg="Node %s not found" % (name))

        for i in range(len(nodes)):
            elem.send_keys(Keys.ARROW_DOWN)
            active = self.driver.find_element_by_css_selector(
                "#flatNodelist li.active").text
            if active == name:
                break
        elem.send_keys(Keys.RETURN)
        elem.send_keys(Keys.BACK_SPACE * len(name))

    def delNode(self, nodeId):
        """Delete a node."""
        self.assertNodeExists(nodeId)
        self.driver.execute_script("""return removeNode("%s");""" % (nodeId))
        time.sleep(0.8)
        self.assertNodeNotExists(nodeId)

    def selectNode(self, nodeid):
        """Select a node. Since it emulates the clicking node,
        you will fail if the nodes are overlapped.
        For such a case, use reshape method in advance.
        """
        el = self.driver.find_element_by_id(nodeid)
        el.click()
        self.assertNodeSelected(nodeid)

    def setParamType(self, nodeId, paramname, setType, isManual=False):
        """Set parameter type."""
        if not isManual:
            self.openParam(nodeId)
        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        xpath += '/descendant::td[@class="type"]'
        xpath += '/*[@name="%s_type"]' % (paramname)

        # print "--" + setType + "--" + elem.get_attribute("value") + "--"
        while 1:
            elem = self.driver.find_element_by_xpath(xpath)
            if elem.get_attribute("value") == setType:
                break
            elem.send_keys(setType[0])

        self.assertParamType(nodeId, paramname, setType)

        if not isManual:
            self.closeParam(nodeId)

    def setParamValue(self, nodeId, paramname, value, isManual=False):
        """Set parameter value."""
        if not isManual:
            self.openParam(nodeId)
        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        xpath += '/descendant::td[@class="value"]'
        xpath += '/*[@name="%s"]' % (paramname)
        elem = self.driver.find_element_by_xpath(xpath)
        elem.send_keys(Keys.BACK_SPACE * 100)
        elem.send_keys(str(value))

        if not isManual:
            self.closeParam(nodeId)

    def getParamValue(self, nodeId, paramname):
        """Get parameter value."""
        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        xpath += '/descendant::td[@class="value"]'
        xpath += '/*[@name="%s"]' % (paramname)
        elem = self.driver.find_element_by_xpath(xpath)
        return elem.get_attribute("value")

    def openParam(self, nodeId):
        """Open parameters dialog of the specified node.
        Since it emulates the clicking node,
        you will fail if the nodes are overlapped.
        For such a case, use reshape method in advance.
        """
        action = webdriver.ActionChains(self.driver)
        action.double_click(self.driver.find_element_by_id(nodeId))
        action.perform()
        time.sleep(0.8)

        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        elem = self.driver.find_element_by_xpath(xpath)
        self.assertTrue(elem.is_displayed(),
                        msg="parameter dialog is not opened")

    def closeParam(self, nodeId, buttonType="ok"):
        """Close the opened parametrs dialog.
        if buttonType = "ok", the ok button is clicked.
        elif buttonType = "close", the close button (upper right) is clicked.
        elif buttonType = "cancel", the cancel button is clicked."""

        xpath = '//div[@id="%s%s%s"]/..' % (nodeId, self.delim, "parameters")
        if buttonType == "close":
            xpath += '/descendant::button[@title="close"]'
        elif buttonType == "ok":
            xpath += '/descendant::div[@class="ui-dialog-buttonset"]'
            xpath += '/button[1]'
        elif buttonType == "cancel":
            xpath += '/descendant::div[@class="ui-dialog-buttonset"]'
            xpath += '/button[2]'

        elem = self.driver.find_element_by_xpath(xpath)
        elem.click()
        time.sleep(0.8)

        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        elem = self.driver.find_element_by_xpath(xpath)
        self.assertFalse(elem.is_displayed(),
                         msg="parameter dialog is not closed")

    def addSheet(self, name, typ="subnet"):
        """Add a new sheet as a subnet."""
        self.driver.find_element_by_id("addSheetButton").click()
        form = self.driver.find_element_by_id("addSheetName")
        form.send_keys(name)

        typeform = self.driver.find_element_by_id("addSheetType")
        typeform.send_keys(typ)

        form.send_keys(Keys.RETURN)
        self.assertTabExists(name)

    def activateSheet(self, name):
        """Select a specified sheet"""
        xpath = '//div[@id="sheets"]'
        xpath += '/ul[@role="tablist"]'
        xpath += '/li[@aria-controls="sheet_%s"]/a' % (name)
        print xpath
        obj = self.driver.find_element_by_xpath(xpath)
        print obj.text

        obj.click()

        self.assertActiveSheet(name)

    def addTerminalRole(self, nodeid, name, roletype, rolename=""):
        """Add a new terminal role to the specified nodeid, and terminal name.
        roletype must be Input/Output/Condition
        rolename can be any name."""

        if not(roletype in ["Input", "Output", "Condition"]):
            raise exceptions.ValueError(
                "roletype must be Input/Output/Condition")

        selector = "#%s%s%s .terminalcircle" % (nodeid, self.delim, name)
        elem = self.driver.find_element_by_css_selector(selector)
        self.driver.execute_script(
            """
            var sheet = _.findObject(network.nodes, "id", "%s").sheet;
            adjustScroll(sheet, function(n){
                   return n.id === "%s";})""" % (nodeid, nodeid))
        time.sleep(0.4)
        elem.click()

        # get context menu
        # for selenium mode, context menu opens with left click
        # action = webdriver.ActionChains(self.driver)
        # action.click(elem).perform()

        # Set as roletype
        elem = self.driver.execute_script(
            "return $('.context-menu-item:visible');")
        for e in elem:
            text = e.text
            if text == "Set as " + roletype:
                e.click()
                break
        time.sleep(0.4)

        # type name if roletype = Input or Output
        if roletype in ["Input", "Output"]:
            elem = self.driver.find_element_by_xpath(
                '//*[@id="terminalRoleName"]')
            if rolename:
                elem.send_keys(Keys.BACK_SPACE * 100)
                elem.send_keys(rolename)
            elem.send_keys(Keys.RETURN)

            if rolename != "":
                self.assertTerminalRoleName(nodeid, name, rolename)
        else:
            print "roletype is Condition"

    def deleteTerminalRole(self, nodeid, name, roletype):
        """Delete the specified terminal role"""
        if not(roletype in ["Input", "Output", "Condition"]):
            raise exceptions.ValueError(
                "roletype must be Input/Output/Condition")

        selector = "#%s%s%s .terminalcircle" % (nodeid, self.delim, name)
        elem = self.driver.find_element_by_css_selector(selector)
        action = webdriver.ActionChains(self.driver)

        # get context menu
        # for selenium mode, context menu opens with left click
        action.click(elem).perform()

        # add Set as something
        for e in self.driver.find_elements_by_class_name("context-menu-item"):
            if not e.is_displayed():
                continue
            if "Delete " + roletype in e.find_element_by_tag_name("span").text:
                e.click()
                break

    def setDisplay(self, name, value):
        """set a praference. currently supported vlaues are:
        Line: Lines / Bezier,
        Language: Japanese / English,
        Show Tooltip : Yes / No"""

        candidates = {"Line": ["Lines", "Bezier"],
                      "Language": ["Japanese", "English"],
                      "Show Tooltip": ["Yes", "No"]}
        table = {"Line": "line",
                 "Language": "lang",
                 "Show Tooltip": "tooltip"}
        if not(name in candidates.keys()):
            raise exceptions.ValueError(
                "name must be " + " / ".join(candidates))

        if not(value in candidates[name]):
            raise exceptions.ValueError(
                "If name = %s, value must be " + " / ".join(candidates[name]))

        # open dialog and select display
        self.driver.find_element_by_id("preferenceButton").click()
        xpath = "//div[@id='preferenceDialog']"
        xpath += "/div[@class='dialogMenu']"
        xpath += "/descendant::button[2]"
        self.driver.find_element_by_xpath(xpath).click()

        # set value
        xpath = "//div[@id='preferenceDialog']"
        xpath += "/descendant::select[@name='%s']" % (table[name])
        self.driver.find_element_by_xpath(xpath).send_keys(value)

        # close dialog
        xpath = "//div[@id='preferenceDialog']/.."
        xpath += "/descendant::button[@title='close']"
        self.driver.find_element_by_xpath(xpath).click()

    def loadFile(self, filetype, filename):
        self.driver.find_element_by_id("fileManagerButton").click()
        el = self.driver.find_element_by_css_selector(
            "#file_%s button.load[name='%s']" % (filetype, filename))
        el.click()

    def reshape(self):
        """Reshapes all nodes. Graphviz is required."""
        self.driver.find_element_by_id("reshapeButton").click()
        time.sleep(1)

    def getTerminalRoleName(self, nodeid, termname):
        """Get a role name of the terminal specified by nodeid and termname"""
        termid = nodeid + self.delim + termname
        # print "assertNodeExists:", nodeid
        # self.assertNodeExists(nodeid)
        # print "assertTermExists:", nodeid, termname
        # self.assertTermExists(nodeid, termname)

        # return self.driver.find_element_by_css_selector(
        #     "#" + termid + " .terminalRoleName").text
        xpath = '/div[@id="%s"]//div/span' % termid
        print xpath
        obj = self.driver.find_element_by_xpath(xpath)
        txt = obj.text
        print txt
        return txt

    def zoom(self, zoomValue):
        """ Chagne the zoom of the network """
        self.assertIn(zoomValue, range(1, 11),
                      msg="zoomValue must be [1,...,10]")
        el = self.driver.find_element_by_xpath("//div[@id='zoomSlider']/a")

        current = self.driver.execute_script(
            'return $("#zoomSlider").slider("value");')

        for i in range(abs(zoomValue - current)):
            el.click()
            if zoomValue < current:
                el.send_keys(Keys.ARROW_LEFT)
            elif zoomValue > current:
                el.send_keys(Keys.ARROW_RIGHT)

    def saveAs(self, name):
        """ Save current network as "name" """
        self.driver.find_element_by_id("save").click()
        el = self.driver.find_element_by_id("networkName")
        el.send_keys(Keys.BACK_SPACE * 100)
        el.send_keys(name)
        el.send_keys(Keys.RETURN)

        if self.driver.switch_to_alert():
            self.driver.switch_to_alert().accept()

        # check if save succeeded
        filename = self.driver.find_element_by_id("currentFileName")
        self.assertEqual(filename.text, name)

        filename = self.driver.execute_script(
            'return app.get("currentFileName")')
        self.assertEqual(filename, name)

    def execute(self):
        """Execute the current network"""
        self.driver.find_element_by_id("execute").click()

    def executeStatus(self):
        """Returns the current execute status"""
        return self.driver.find_element_by_id("executeStatus").text

    def assertNodeExists(self, nodeid):
        """Tests whether the specified node exists."""
        els = self.driver.find_element_by_id(nodeid)
        self.assertEqual(els.get_attribute("id"), nodeid,
                         msg="node %s does not exist" % (nodeid))
        self.assertTrue(self.driver.execute_script(
            'return _.findObject(network.nodes, "id", "%s") !== undefined;' % (
                nodeid)))

    def assertNodeNotExists(self, nodeid):
        """Tests whether the specified node not exists."""
        self.driver.implicitly_wait(0.1)
        els = self.driver.find_elements_by_id(nodeid)
        self.driver.implicitly_wait(self.timeout)
        self.assertEqual(len(els), 0, msg="node %s exists" % (nodeid))

    def assertTermExists(self, nodeid, termname):
        """Tests whether the nodeid has a terminal whose name is termname"""
        els = self.driver.find_element_by_id(nodeid + self.delim + termname)
        print els
        self.assertEqual(
            els.get_attribute("id"), nodeid + self.delim + termname,
            msg="node %s does not have a terminal %s" % (nodeid, termname))

    def assertTermNotExists(self, nodeid, termname):
        """Tests whether the nodeid does not have
         a terminal whose name is termname"""
        self.driver.implicitly_wait(0.1)
        els = self.driver.find_elements_by_id(
            nodeid + self.delim + termname)
        self.driver.implicitly_wait(self.timeout)
        self.assertEqual(
            len(els), 0,
            msg="node %s havs a terminal %s" % (nodeid, termname))

    def assertNodeSelected(self, nodeid):
        """Tests whether the specified node is selected."""
        time.sleep(0.3)
        elem = self.driver.find_element_by_id(nodeid)
        self.assertIn("nodeSelected", elem.get_attribute("class"),
                      msg="node %s is not selected." % (nodeid))

    def assertActiveSheet(self, sheetname):
        """Tests whether the active sheet name is sheetname"""
        xpath = '//div[@id="sheets"]'
        xpath += '/descendant::li[@aria-controls="sheet_%s"]/a' % (sheetname)
        activename = self.driver.find_element_by_xpath(xpath).text
        self.assertEqual(
            activename, sheetname,
            msg="Active tab is not %s but %s." % (activename, sheetname))

        self.assertEqual(
            sheetname, self.driver.execute_script("return currentTab;"),
            msg="javascript: currentTab is not %s but %s." % (
                activename, sheetname))

    def assertTabExists(self, sheetname):
        """Tests whether the sheet whose name is sheetname exists."""
        xpath = '//div[@id="sheets"]'
        xpath += '/descendant::li[@role="tab"]'
        tabs = map(lambda e: e.get_attribute("aria-controls"),
                   self.driver.find_elements_by_xpath(xpath))
        self.assertIn("sheet_" + sheetname, tabs,
                      msg="sheet %s does not exist." % (sheetname))

        self.assertTrue(self.driver.execute_script(
            """return _.findObject(network.sheets, "name", "%s");""" % (
                sheetname)),
            msg="javascript: sheet %s is not added" % (sheetname))

    def assertTerminalRoleName(self, nodeId, termname, rolename):
        """Tests whether the terminal's role name is rolename"""
        gotRolename = self.getTerminalRoleName(nodeId, termname)
        self.assertEqual(gotRolename, rolename)

        ret = self.driver.execute_script(
            """return _.find(network.terminalRoles, function(r){
                return r.nodeId === "%s" &&
                       r.termName === "%s" &&
                       r.name === "%s";
                }) !== undefined;""" % (nodeId, termname, rolename))
        self.assertTrue(ret, msg="javascript: terminal role is not found")

    def assertParamType(self, nodeId, paramname, typ):
        """Tests whether the parameter has the specified type"""
        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        xpath += '/descendant::td[@class="type"]'
        xpath += '/*[@name="%s_type"]' % (paramname)
        elem = self.driver.find_element_by_xpath(xpath)
        self.assertEqual(typ, elem.get_attribute("value"))

        objtyp = self.driver.execute_script(
            """var node = _.find(network.nodes,
                                 function(n){return n.id === "%s";});
               var param = _.find(node.parameters,
                                  function(p){return p.name === "%s";});
               return param.type;""" % (nodeId, paramname))
        self.assertEqual(
            objtyp, typ,
            msg="The type of Param %s for node %s should %s but %s " % (
                paramname, nodeId, typ, objtyp))

    def assertNotParamType(self, nodeId, paramname, typ):
        """Tests whether the parameter has the specified type"""
        xpath = '//form[@id="%s%s%s"]' % (nodeId, self.delim, "parametersForm")
        xpath += '/descendant::td[@class="type"]'
        xpath += '/*[@name="%s_type"]' % (paramname)
        elem = self.driver.find_element_by_xpath(xpath)
        self.assertEqual(typ, elem.get_attribute("value"))

        objtyp = self.driver.execute_script(
            """var node = _.find(network.nodes,
                                 function(n){return n.id === "%s";});
               var param = _.find(node.parameters,
                                  function(p){return p.name === "%s";});
               return param.type;""" % (nodeId, paramname))
        self.assertNotEqual(objtyp, typ)

    def assertParamValue(self, nodeId, paramname, val):
        """Tests whether the parameter has the specified value"""
        objval = self.driver.execute_script(
            """var node = _.find(network.nodes,
                                 function(n){return n.id === "%s";});
               var param = _.find(node.parameters,
                                  function(p){return p.name === "%s";});
               return param.value;""" % (nodeId, paramname))
        self.assertEqual(
            str(objval), str(val),
            msg="The parameter %s of nodeid %s should %s but %s" % (
                paramname, nodeId, val, objval))

    def assertNotParamValue(self, nodeId, paramname, val):
        """Tests whether the parameter has the specified value"""
        objval = self.driver.execute_script(
            """var node = _.find(network.nodes,
                                 function(n){return n.id === "%s";});
               var param = _.find(node.parameters,
                                  function(p){return p.name === "%s";});
               return param.value;""" % (nodeId, paramname))
        self.assertNotEqual(
            str(objval), str(val),
            msg="The parameter %s of nodeid %s should not %s but %s" % (
                paramname, nodeId, val, objval))

    def assertParamValueError(self):
        """Tests whether the parameter error is occured"""
        self.assertNotEqual(
            "",
            self.driver.find_element_by_id("executeStatus").text)

    def screenshot(self, dirname=".", prefix="", fullscreen=False,
                   doNothing=False, doReshape=True):
        """Takes a screenshot."""
        if fullscreen and not doNothing:
            self.driver.execute_script("forScreenshot()")
        elif not fullscreen and not doNothing:
            self.driver.find_element_by_id("minimizeNodelistButton").click()

        if doReshape:
            self.reshape()

        sheets = self.driver.execute_script("return _.keys(network.sheets);")
        if len(sheets) == 1:
            self.driver.save_screenshot("%s/%s%s.png" % (
                dirname, prefix, sheets[0]))
        else:
            for sheet in sheets:
                self.activateSheet(sheet)
                self.driver.save_screenshot("%s/%s%s.png" % (
                    dirname, prefix, sheet))
